/* Copyright (c) 2010 Dovecot authors, see the included COPYING file */

/*
   eval `cat /usr/local/lib/dovecot/dovecot-config`
   gcc -fPIC -shared -DHAVE_CONFIG_H \
     `echo $DOVECOT_CFLAGS $LIBDOVECOT_INCLUDE $LIBDOVECOT_STORAGE_INCLUDE` \
     mail-filter-plugin.c istream-ext-filter.c -o mail_filter_plugin.so
*/

#include "lib.h"
#include "array.h"
#include "str.h"
#include "safe-mkstemp.h"
#include "close-keep-errno.h"
#include "mail-user.h"
#include "mail-storage-private.h"
#include "istream.h"
#include "istream-seekable.h"
#include "istream-ext-filter.h"

/* After buffer grows larger than this, create a temporary file to /tmp
   where to read the mail. */
#define MAIL_MAX_MEMORY_BUFFER (1024*128)

#define MAIL_FILTER_MAIL_CONTEXT(obj) \
	MODULE_CONTEXT(obj, mail_filter_mail_module)

const char *mail_filter_plugin_version = DOVECOT_VERSION;

static MODULE_CONTEXT_DEFINE_INIT(mail_filter_mail_module,
				  &mail_module_register);

static int seekable_fd_callback(const char **path_r, void *context)
{
	struct mail_user *user = context;
	string_t *path;
	int fd;

	path = t_str_new(128);
	mail_user_set_get_temp_prefix(path, user->set);
	fd = safe_mkstemp(path, 0600, (uid_t)-1, (gid_t)-1);
	if (fd == -1) {
		i_error("safe_mkstemp(%s) failed: %m", str_c(path));
		return -1;
	}

	/* we just want the fd, unlink it */
	if (unlink(str_c(path)) < 0) {
		/* shouldn't happen.. */
		i_error("unlink(%s) failed: %m", str_c(path));
		close_keep_errno(fd);
		return -1;
	}

	*path_r = str_c(path);
	return fd;
}

static int
mail_filter_istream_opened(struct mail *_mail, struct istream **stream)
{
	struct mail_private *mail = (struct mail_private *)_mail;
	union mail_module_context *mmail = MAIL_FILTER_MAIL_CONTEXT(mail);
	struct istream *input, *inputs[2];
	const char *args;

	args = mail_user_plugin_getenv(_mail->box->storage->user,
				       "mail_filter_executable");
	if (args != NULL) {
		input = *stream;
		*stream = i_stream_create_ext_filter(input, args);
		i_stream_unref(&input);

		inputs[0] = *stream;
		inputs[1] = NULL;
		*stream = i_stream_create_seekable(inputs, MAIL_MAX_MEMORY_BUFFER,
						   seekable_fd_callback,
						   _mail->box->storage->user);
		i_stream_unref(&inputs[0]);
	}

	return mmail->super.istream_opened(_mail, stream);
}

static void mail_filter_mail_allocated(struct mail *_mail)
{
	struct mail_private *mail = (struct mail_private *)_mail;
	struct mail_vfuncs *v = mail->vlast;
	union mail_module_context *mmail;

	mmail = p_new(mail->pool, union mail_module_context, 1);
	mmail->super = *v;
	mail->vlast = &mmail->super;

	v->istream_opened = mail_filter_istream_opened;
	MODULE_CONTEXT_SET_SELF(mail, mail_filter_mail_module, mmail);
}

static struct mail_storage_hooks mail_filter_mail_storage_hooks = {
	.mail_allocated = mail_filter_mail_allocated
};

void mail_filter_plugin_init(struct module *module);
void mail_filter_plugin_deinit(void);

void mail_filter_plugin_init(struct module *module)
{
	mail_storage_hooks_add(module, &mail_filter_mail_storage_hooks);
}

void mail_filter_plugin_deinit(void)
{
	mail_storage_hooks_remove(&mail_filter_mail_storage_hooks);
}
